from collections import (
    defaultdict,
    deque,
)
from typing import (  # noqa: F401
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from .base import (  # noqa: F401
    BaseDocstring,
    DocstringStyle,
    Sections,
)
from ..config import Configuration
from ..node import (
    CykNode,
)
from ..parse.google import (
    parse,
)
from ..lex import (
    condense,
    lex,
)
from ..errors import (
    DarglintError,
)
from ..config import (
    Strictness,
)
from ..parse.identifiers import (
    ArgumentIdentifier,
    ArgumentItemIdentifier,
    ArgumentTypeIdentifier,
    ExceptionIdentifier,
    ExceptionItemIdentifier,
    Identifier,
    NoqaIdentifier,
)


class _CykVisitor(object):
    """A class for walking the CYK tree.

    This visitor can mark branches of the tree with an arbitrary
    string.  This is useful for giving context to leaf nodes which
    must be interpreted in the context of their branch root.

    """

    def __init__(self, root):
        self.stack = deque([root])
        self.marks = deque([])

    def __iter__(self):
        return self

    def __next__(self):
        if self.stack:
            curr = self.stack.pop()
            if self.marks:
                mark = self.marks.pop()
            else:
                mark = None
            if curr.lchild:
                self.stack.append(curr.lchild)
                if mark:
                    self.marks.append(mark)
            if curr.rchild:
                self.stack.append(curr.rchild)
                if mark:
                    self.marks.append(mark)
            return (curr, mark)
        else:
            raise StopIteration()

    def mark(self, mark):
        assert not self.marks, 'Marks should be non-overlapping.'
        self.marks.append(mark)


class Docstring(BaseDocstring):
    """The docstring class interprets the AST of a docstring."""

    def __init__(self, root, style=DocstringStyle.GOOGLE, config=None):
        # type: (Union[CykNode, str], DocstringStyle, Optional[Configuration]) -> None
        """Create a new docstring from the AST.

        Args:
            root: The root of the AST, or the docstring
                (as a string.)  If it is a string, the
                string will be parsed.
            style: The style of the docstring.  Discarded,
                since this Docstring is always the Google style.
            config: The configuration parameters.  Passed to
                lex to change the number of spaces which count
                as an indentation.

        """
        if isinstance(root, CykNode):
            self.root = root
        else:
            self.root = parse(condense(lex(root, config)))
        self._lookup = self._discover()

    def _discover(self):
        # type: () -> Dict[str, List[CykNode]]
        """Walk the tree, finding all non-terminal nodes.

        Returns:
            A lookup table for compound Nodes by their NodeType.

        """
        lookup = defaultdict(
            lambda: list()
        )  # type: Dict[str, List[CykNode]]
        for node in self.root.in_order_traverse():
            if node.annotations:
                for annotation in node.annotations:
                    if issubclass(annotation, Identifier):
                        lookup[annotation.key].append(node)
            lookup[node.symbol].append(node)
        return lookup

    def get_section(self, section):
        # type: (Sections) -> Optional[str]
        nodes = []  # type: Optional[List[CykNode]]

        if section == Sections.SHORT_DESCRIPTION:
            nodes = self._lookup.get('short-description', None)
        elif section == Sections.LONG_DESCRIPTION:
            nodes = self._lookup.get('long-description', None)
        elif section == Sections.ARGUMENTS_SECTION:
            nodes = self._lookup.get('arguments-section', None)
        elif section == Sections.RAISES_SECTION:
            nodes = self._lookup.get('raises-section', None)
        elif section == Sections.YIELDS_SECTION:
            nodes = self._lookup.get('yields-section', None)
        elif section == Sections.RETURNS_SECTION:
            nodes = self._lookup.get('returns-section', None)
        elif section == Sections.NOQAS:
            nodes = self._lookup.get('noqa', None)
        else:
            raise Exception(
                'Unsupported section type {}'.format(section.name)
            )

        if not nodes:
            return None

        return_value = ''
        for node in nodes:
            return_value += '\n\n' + node.reconstruct_string()

        return return_value.strip() or None

    def _get_argument_types(self):
        # type: () ->  Optional[List[Optional[str]]]
        """Get a list of types corresponding to arguments.

        Returns:
            A dictionary matching arguments to types.

        """
        lookup = self._get_compound_item_type_lookup('arguments-section')
        if not lookup:
            return None

        names_and_types = sorted(lookup.items())
        return [x[1] for x in names_and_types]

    def _get_return_type(self):
        # type: () -> Optional[str]
        """Get the return type specified by the docstring, if any.

        Returns:
            The return type or None.

        """
        for type_node in self._lookup['returns-type']:
            if not type_node.lchild or not type_node.lchild.value:
                continue
            return type_node.lchild.value.value
        return None

    def get_types(self, section):
        # type: (Sections) -> Union[None, str, List[Optional[str]]]
        if section == Sections.ARGUMENTS_SECTION:
            if 'arguments-section' not in self._lookup:
                return None
            return self._get_argument_types()
        elif section == Sections.RETURNS_SECTION:
            return self._get_return_type()
        elif section == Sections.YIELDS_SECTION:
            return self._get_yield_type()
        else:
            raise Exception(
                'Section type {} does not have types, '.format(
                    section.name
                ) + 'or is not yet supported'
            )
        return None

    def _get_compound_item_type_lookup(self, node_type):
        # type: (str) -> Optional[Dict[str, Optional[str]]]
        """Get a map of names to types for the section.

        Args:
            node_type: The type of the section.

        Returns:
            A lookup of items to types.

        """
        if node_type not in self._lookup:
            return None

        item_types = dict()  # type: Dict[str, Optional[str]]
        for item in self._lookup[ArgumentTypeIdentifier.key]:
            item_types[
                ArgumentIdentifier.extract(item)
            ] = ArgumentTypeIdentifier.extract(item)
        for item in self._lookup[ArgumentIdentifier.key]:
            name = ArgumentIdentifier.extract(item)
            if name not in item_types:
                item_types[name] = None
        return item_types

    def _get_compound_items(self, symbol):
        # type: (str) -> Optional[List[str]]
        lookup = self._get_compound_item_type_lookup(symbol)
        if lookup is None:
            return None
        return sorted(lookup.keys())

    def _get_raises_section_items(self):
        # type: () -> Optional[List[str]]
        items = list()
        for item in self._lookup[ExceptionIdentifier.key]:
            items.append(ExceptionIdentifier.extract(item))
        return sorted(items) or None

    def get_items(self, section):
        # type: (Sections) -> Optional[List[str]]
        if section == Sections.ARGUMENTS_SECTION:
            return self._get_compound_items('arguments-section')
        elif section == Sections.RAISES_SECTION:
            return self._get_raises_section_items()
        else:
            raise Exception(
                'Section type {} does not have items, '.format(
                    section.name
                ) + 'or is not yet supported.'
            )
        return None

    def _get_yield_type(self):
        # type: () -> Optional[str]
        """Get the yield type specified by the docstring, if any.

        Returns:
            The yield type or None

        """
        for type_node in self._lookup['yields-type']:
            if not type_node.lchild or not type_node.lchild.value:
                continue
            return type_node.lchild.value.value
        return None

    def get_noqas(self):
        # type: () -> Dict[str, List[str]]
        """Get a map of the errors ignored to their targets.

        Returns:
            A dictionary containing the errors to ignore as keys and
            a list of which targets to apply these exceptions to as
            the values.  A blank list implies a global noqa.

        """
        def _get_branch_name(node):
            for child in node.walk():
                for annotation in child.annotations:
                    if (
                        issubclass(annotation, ArgumentIdentifier)
                        or issubclass(annotation, ExceptionIdentifier)
                    ):
                        return annotation.extract(child)

        noqas = defaultdict(lambda: set())  # type: Dict[str, Set[str]]

        for node in self._lookup[NoqaIdentifier.key]:
            error = NoqaIdentifier.extract(node)
            if error:
                noqas[error] |= set(NoqaIdentifier.extract_targets(node))
            else:
                noqas['*'] = set()

        # If the noqa did not have a target, but it had an error and was
        # in an item or error description, then the item or error in question
        # is the target.
        for section in (
            self._lookup['arguments-section'] + self._lookup['raises-section']
        ):
            visitor = _CykVisitor(section)
            for node, mark in visitor:
                for annotation in node.annotations:
                    if issubclass(annotation, ArgumentItemIdentifier):
                        visitor.mark(_get_branch_name(node))
                    elif issubclass(annotation, ExceptionItemIdentifier):
                        visitor.mark(_get_branch_name(node))
                    elif issubclass(annotation, NoqaIdentifier):
                        error = annotation.extract(node)
                        if error and mark:
                            noqas[error].add(mark)

        for section in self._lookup['raises-section']:
            pass

        return {
            key: sorted(values)
            for key, values in noqas.items()
        }

    def get_style_errors(self):
        # type: () -> Iterable[Tuple[Callable, Tuple[int, int]]]
        """Get any style errors annotated on the tree.

        Yields:
            Instances of DarglintErrors for style issues.

        # noqa: I302

        """
        for node, _ in _CykVisitor(self.root):
            for annotation in node.annotations:
                if issubclass(annotation, DarglintError):
                    yield annotation, node.line_numbers

    def get_line_numbers(self, symbol):
        # type: (str) -> Optional[Tuple[int, int]]
        """Get the line numbers for the first instance of the given section.

        Args:
            symbol: The type of node which we want line numbers for.
                These should be unique instances. (I.e. they should be
                in the set of compound NodeTypes which only occur
                once in a docstring. For example, "raises-section" and
                "arguments-section".

        Returns:
            The line numbers for the first instance of the given node type.

        """
        nodes = self._lookup[symbol]
        if nodes:
            return nodes[0].line_numbers
        return None

    def get_line_numbers_for_value(self, symbol, value):
        # type: (str, str) -> Optional[Tuple[int, int]]
        """Get the line number for a node with the given value.

        Args:
            symbol: The compound node which should contain the
                node we are searching for.
            value: The value of the node.

        Returns:
            A list of line numbers for nodes which match the
            parameters.

        """
        nodes = self._lookup[symbol]
        for node in nodes:
            for child in node.walk():
                if (child.value
                        and child.value.value == value
                        and child.line_numbers):
                    return child.line_numbers
        return None

    @property
    def ignore_all(self):
        # type: () -> bool
        """Return whether we should ignore everything in the docstring.

        This happens when there is a bare noqa in the docstring, or
        there is "noqa: *" in the docstring.

        Returns:
            True if we should ignore everything, otherwise false.

        """
        noqas = self.get_noqas()
        return '*' in noqas

    def satisfies_strictness(self, strictness):
        # type: (Strictness) -> bool
        """Return true if the docstring has no more than the min strictness.

        Args:
            strictness: The minimum amount of strictness which should
                be present in the docstring.

        Returns:
            True if there is no more than the minimum amount of strictness.

        """
        sections = {
            section for section in Sections
            for section in {
                Sections.SHORT_DESCRIPTION,
                Sections.LONG_DESCRIPTION,
                Sections.ARGUMENTS_SECTION,
                Sections.RAISES_SECTION,
                Sections.YIELDS_SECTION,
                Sections.RETURNS_SECTION,
                Sections.NOQAS,
            }
            if self.get_section(section)
        }
        if strictness == Strictness.SHORT_DESCRIPTION:
            return sections == {Sections.SHORT_DESCRIPTION}
        elif strictness == Strictness.LONG_DESCRIPTION:
            return sections in [
                {Sections.SHORT_DESCRIPTION},
                # Shouldn't be possible, but if it is in the future, then
                # we should allow this.
                {Sections.LONG_DESCRIPTION},
                {Sections.SHORT_DESCRIPTION, Sections.LONG_DESCRIPTION},
            ]
        else:
            return False



import inspect
from typing import (
    List,
)
from functools import (
    reduce,
)

from ..token import (
    Token,
    TokenType,
    KEYWORDS,
)
from .cyk import (
    parse as cyk_parse,
)
from ..node import (
    CykNode,
)

from .combinator import (
    parser_combinator,
)
from .long_description import (
    parse as long_description_parse,
)

from .grammars.google_arguments_section import ArgumentsGrammar
from .grammars.google_raises_section import RaisesGrammar
from .grammars.google_returns_section import ReturnsGrammar
from .grammars.google_returns_section_without_type import (
    ReturnsWithoutTypeGrammar,
)
from .grammars.google_short_description import ShortDescriptionGrammar
from .grammars.google_yields_section import YieldsGrammar
from .grammars.google_yields_section_without_type import (
    YieldsWithoutTypeGrammar,
)


def _get_split_end_with_indents(tokens, i):
    # type: (List[Token], int) -> int
    """Return the index of the end of this split, or 0.

    Args:
        tokens: A list of tokens.
        i: The current index.

    Returns:
        If i is the start of a split, return the index of the
        token after the end of the split (or the last token, if
        it's the end of the docstring.)  If we're not at a split,
        return 0.

    """
    newline_count = 0
    newline_run = 0
    highest_newline_run = 0
    j = i
    while j < len(tokens):
        if tokens[j].token_type == TokenType.NEWLINE:
            newline_count += 1
            newline_run += 1
            if newline_run > highest_newline_run:
                highest_newline_run = newline_run
        elif tokens[j].token_type == TokenType.INDENT:
            newline_run = 0
        else:
            break
        j += 1

    # Back up so that we don't remove indents on the same line as
    # the encountered text.
    while (j < len(tokens)
            and j > 1
            and tokens[j - 1].token_type == TokenType.INDENT):
        j -= 1

    # TODO: Do we want to check for keywords before assuming a
    # new section?  If we have line-separated sections in args,
    # which do not have indents, then we will parse incorrectly.
    if newline_count < 2:
        return 0

    # If there are two newlines in a row, we have a break, no
    # matter what.
    if highest_newline_run > 1:
        return j

    # If there were not 2+ newlines in a row, (i.e. there were
    # indented lines in with these), then it's only a new section
    # if it starts with a keyword.
    if (j < len(tokens)
            and tokens[j].token_type in KEYWORDS
            and tokens[j - 1].token_type == TokenType.NEWLINE):
        return j

    return 0


def top_parse(tokens):
    # type: (List[Token]) -> List[List[Token]]
    all_sections = list()
    curr = 0
    # Strip leading newlines.
    while curr < len(tokens) and tokens[curr].token_type == TokenType.NEWLINE:
        curr += 1
    prev = curr

    while curr < len(tokens):
        split_end = _get_split_end_with_indents(tokens, curr)
        if split_end > curr:
            if tokens[prev:curr]:
                all_sections.append(
                    tokens[prev:curr]
                )
            curr = split_end
            prev = curr
        else:
            curr += 1
    last_section = tokens[prev:curr]
    if last_section:
        all_sections.append(last_section)
    return all_sections


def _match(token):
    """Match the given token from the given section to a set of grammars.

    Args:
        token: The token to match.  This should hint at what sections
            could possibly be here.

    Returns:
        A list of grammars to be tried in order.

    """
    tt_lookup = {
        TokenType.RETURNS: [
            ReturnsGrammar,
            ReturnsWithoutTypeGrammar,
            long_description_parse,
        ],
        TokenType.ARGUMENTS: [
            ArgumentsGrammar,
            long_description_parse,
        ],
        TokenType.YIELDS: [
            YieldsGrammar,
            YieldsWithoutTypeGrammar,
            long_description_parse,
        ],
        TokenType.RAISES: [
            RaisesGrammar,
            long_description_parse,
        ],
    }
    return tt_lookup.get(token.token_type, [long_description_parse])


def lookup(section, section_index=-1):
    assert len(section) > 0
    grammars = _match(section[0])
    if section_index == 0:
        return [ShortDescriptionGrammar] + grammars
    return grammars


def combinator(*args):
    def inner(*nodes):
        if len(nodes) == 1:
            return CykNode(
                symbol='docstring',
                lchild=nodes[0],
            )
        elif len(nodes) == 2:
            return CykNode(
                symbol='docstring',
                lchild=nodes[0],
                rchild=nodes[1],
            )
    if args:
        return reduce(inner, args)
    else:
        # The arguments are empty, so we return an
        # empty docstring.
        return CykNode(symbol='docstring')


def parse(tokens):
    def mapped_lookup(section, section_index=-1):
        for grammar in lookup(section, section_index):
            if inspect.isclass(grammar):
                yield lambda x: cyk_parse(grammar, x)
            else:
                yield grammar
    return parser_combinator(top_parse, mapped_lookup, combinator, tokens)



# Generated on 2020-04-14 07:01:35.379938

from darglint.parse.grammar import (
    BaseGrammar,
    P,
)

from darglint.token import (
    TokenType,
)

from darglint.parse.identifiers import (
    NoqaIdentifier,
)

from darglint.errors import (
    EmptyDescriptionError,
    EmptyTypeError,
    IndentError,
)

from darglint.parse.identifiers import (
    ArgumentIdentifier,
    ArgumentItemIdentifier,
    ArgumentTypeIdentifier,
)

class ArgumentsGrammar(BaseGrammar):
    productions = [
        P("arguments-section", ([], "arguments", "arguments-section1", 0)),
        P("items-argument", ([], "item-argument", "items-argument0", 0), ([ArgumentItemIdentifier], "head-argument", "item-body", 0), ([ArgumentIdentifier, EmptyDescriptionError], "indent", "head-argument1", 0), ([ArgumentIdentifier, EmptyTypeError, EmptyDescriptionError], "indent", "head-argument3", 2), ([ArgumentIdentifier, ArgumentTypeIdentifier, EmptyDescriptionError], "indent", "head-argument7", 0), ([ArgumentIdentifier, EmptyDescriptionError], "indent", "head-argument10", 0)),
        P("item-argument", ([ArgumentItemIdentifier], "head-argument", "item-body", 0), ([ArgumentIdentifier, EmptyDescriptionError], "indent", "head-argument1", 0), ([ArgumentIdentifier, EmptyTypeError, EmptyDescriptionError], "indent", "head-argument3", 2), ([ArgumentIdentifier, ArgumentTypeIdentifier, EmptyDescriptionError], "indent", "head-argument7", 0), ([ArgumentIdentifier, EmptyDescriptionError], "indent", "head-argument10", 0)),
        P("head-argument", ([ArgumentIdentifier], "indent", "head-argument1", 0), ([ArgumentIdentifier, EmptyTypeError], "indent", "head-argument3", 2), ([ArgumentIdentifier, ArgumentTypeIdentifier], "indent", "head-argument7", 0), ([ArgumentIdentifier], "indent", "head-argument10", 0)),
        P("item-body", ([], "line", "item-body0", 2), ([], "line", "item-body1", 2), ([], "line", "newline", 2), ([], "word", "line", 2), ([], "word", "noqa-maybe", 2), ([NoqaIdentifier], "hash", "noqa", 2), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 2), (TokenType.INDENT, 2), (TokenType.COLON, 2), (TokenType.HASH, 2), (TokenType.LPAREN, 2), (TokenType.RPAREN, 2), (TokenType.WORD, 2), (TokenType.RAISES, 2), (TokenType.ARGUMENTS, 2), (TokenType.ARGUMENT_TYPE, 2), (TokenType.RETURNS, 2), (TokenType.RETURN_TYPE, 2), (TokenType.YIELDS, 2), (TokenType.YIELD_TYPE, 2), (TokenType.VARIABLES, 2), (TokenType.VARIABLE_TYPE, 2), (TokenType.NOQA, 2), (TokenType.OTHER, 2), (TokenType.RECEIVES, 2), (TokenType.WARNS, 2), (TokenType.SEE, 2), (TokenType.ALSO, 2), (TokenType.NOTES, 2), (TokenType.EXAMPLES, 2), (TokenType.REFERENCES, 2), (TokenType.HEADER, 2), ([IndentError], "line", "item-body4", 0), ([IndentError], "line", "item-body6", 0)),
        P("paragraph-indented-two", ([], "indented-two", "paragraph-indented-two0", 0), ([], "indented-two", "line", 0)),
        P("paragraph", ([], "indents", "paragraph0", 0), ([], "indents", "line", 0), ([], "line", "paragraph2", 0), ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0), ([], "line", "paragraph1", 0)),
        P("line", ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("indented-two", ([], "indent", "indented-two0", 0)),
        P("indents", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("newlines", ([], "newline", "newlines", 0), (TokenType.NEWLINE, 0)),
        P("word", (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("ident", (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0)),
        P("arguments", (TokenType.ARGUMENTS, 0)),
        P("colon", (TokenType.COLON, 0)),
        P("hash", (TokenType.HASH, 0)),
        P("indent", (TokenType.INDENT, 0)),
        P("lparen", (TokenType.LPAREN, 0)),
        P("newline", (TokenType.NEWLINE, 0)),
        P("rparen", (TokenType.RPAREN, 0)),
        P("noqa", (TokenType.NOQA, 0)),
        P("noqa-maybe", ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0)),
        P("noqa-head", ([], "hash", "noqa", 0)),
        P("words", ([], "word", "words", 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("type-section-parens", ([], "lparen", "type-section-parens0", 0)),
        P("type-words-colon", ([], "type-word-colon", "type-words-colon", 0), ([], "type-word-colon", "type-words-colon0", 0), (TokenType.COLON, 0), (TokenType.INDENT, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0)),
        P("type-word-colon", (TokenType.COLON, 0), (TokenType.INDENT, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0)),
        P("arguments-section1", ([], "colon", "arguments-section2", 0)),
        P("arguments-section2", ([], "newline", "arguments-section3", 0)),
        P("arguments-section3", ([], "items-argument", "newlines", 0), ([], "item-argument", "items-argument0", 0), ([ArgumentItemIdentifier], "head-argument", "item-body", 0), ([ArgumentIdentifier, EmptyDescriptionError], "indent", "head-argument1", 0), ([ArgumentIdentifier, EmptyTypeError, EmptyDescriptionError], "indent", "head-argument3", 2), ([ArgumentIdentifier, ArgumentTypeIdentifier, EmptyDescriptionError], "indent", "head-argument7", 0), ([ArgumentIdentifier, EmptyDescriptionError], "indent", "head-argument10", 0)),
        P("items-argument0", ([], "newline", "items-argument", 0)),
        P("head-argument1", ([], "ident", "colon", 0)),
        P("head-argument3", ([], "ident", "head-argument4", 0)),
        P("head-argument4", ([], "lparen", "head-argument5", 0)),
        P("head-argument5", ([], "rparen", "colon", 0)),
        P("head-argument7", ([], "ident", "head-argument8", 0)),
        P("head-argument8", ([], "type-section-parens", "colon", 0)),
        P("head-argument10", ([], "ident", "head-argument11", 0)),
        P("head-argument11", ([], "colon", "newline", 0)),
        P("item-body0", ([], "newline", "paragraph-indented-two", 0)),
        P("item-body1", ([], "newline", "item-body2", 0)),
        P("item-body2", ([], "paragraph-indented-two", "newline", 0)),
        P("item-body4", ([], "newline", "paragraph", 0)),
        P("item-body6", ([], "newline", "item-body7", 0)),
        P("item-body7", ([], "paragraph", "newline", 0)),
        P("paragraph-indented-two0", ([], "line", "paragraph-indented-two1", 0)),
        P("paragraph-indented-two1", ([], "newline", "paragraph-indented-two", 0)),
        P("paragraph0", ([], "line", "paragraph1", 0)),
        P("paragraph1", ([], "newline", "paragraph", 0)),
        P("paragraph2", ([], "newline", "paragraph", 0)),
        P("indented-two0", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("noqa-statement1", ([], "colon", "words", 0)),
        P("type-section-parens0", ([], "type-words-colon", "rparen", 0), (TokenType.RPAREN, 0)),
        P("type-words-colon0", ([], "newline", "type-words-colon1", 0), (TokenType.NEWLINE, 0)),
        P("type-words-colon1", ([], "indents", "type-words-colon", 0), ([], "indent", "indents", 0), (TokenType.INDENT, 0), ([], "type-word-colon", "type-words-colon", 0), ([], "type-word-colon", "type-words-colon0", 0), (TokenType.COLON, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0)),
    ]
    start = "arguments-section"


"""An implementation of the CYK algorithm.

The CYK algorithm was chosen because the Google
docstring format allows for ambiguous representations,
which CYK can handle without devolving into a terrible
complexity. (It has a worst case of O(n^3).

There are faster, on average, algorithms, which might
be better suited to the average task of Darglint.
However, CYK is relatively simple, and is well documented.
(Others, like chart parsing, are much more difficult
to find examples of.)

This representation was based directly on the wikipedia
article, https://en.wikipedia.org/wiki/CYK_algorithm.

"""

from typing import (
    Optional,
    List,
)

from .grammar import (
    BaseGrammar,
)
from ..token import (
    Token,
)
from ..node import (
    CykNode,
)


def parse(grammar, tokens):
    # type: (BaseGrammar, List[Token]) -> Optional[CykNode]
    if not tokens:
        return None
    n = len(tokens)
    r = len(grammar.productions)
    P = [
        [[None for _ in range(r)] for _ in range(n)]
        for _ in range(n)
    ]  # type: List[List[List[Optional[CykNode]]]]
    lookup = grammar.get_symbol_lookup()
    for s, token in enumerate(tokens):
        for v, production in enumerate(grammar.productions):
            for rhs in production.rhs:
                if len(rhs) > 2:
                    continue
                token_type, weight = rhs
                if token.token_type == token_type:
                    P[0][s][v] = CykNode(
                        production.lhs,
                        value=token,
                        weight=weight,
                    )
    for l in range(2, n + 1):
        for s in range(n - l + 2):
            for p in range(l):
                for a, production in enumerate(grammar.productions):
                    for derivation in production.rhs:
                        is_terminal_derivation = len(derivation) <= 2
                        if is_terminal_derivation:
                            continue
                        annotations, B, C, weight = derivation
                        b = lookup[B]
                        c = lookup[C]
                        lchild = P[p - 1][s - 1][b]
                        rchild = P[l - p - 1][s + p - 1][c]
                        if lchild and rchild:
                            old = P[l - 1][s - 1][a]
                            if old and old.weight > weight:
                                continue
                            P[l - 1][s - 1][a] = CykNode(
                                production.lhs,
                                lchild,
                                rchild,
                                annotations=annotations,
                                weight=weight,
                            )
    return P[n - 1][0][lookup[grammar.start]]



from typing import (
    Optional,
    Union,
)

from ..config import Configuration
from .base import BaseDocstring  # noqa
from . import google, sphinx, numpy


class Docstring(object):
    """A factory method for creating docstrings."""

    @staticmethod
    def from_google(root, config=None):
        # type: (str, Optional[Configuration]) -> BaseDocstring
        return google.Docstring(root, config=config)

    @staticmethod
    def from_sphinx(root, config=None):
        # type: (str, Optional[Configuration]) -> BaseDocstring
        return sphinx.Docstring(root, config=config)

    @staticmethod
    def from_numpy(root, config=None):
        # type: (str, Optional[Configuration]) -> BaseDocstring
        return numpy.Docstring(root, config=config)



"""A linter for docstrings following the google docstring format."""
import ast
from collections import deque
import sys
from typing import (
    Callable,
    Iterator,
    List,
    Set,
    Tuple,
    Optional,
    Union,
    Type,
    Any,
)

from .analysis.raise_visitor import (
    RaiseVisitor,
)
from .config import get_logger


logger = get_logger()


FunctionDef = ast.FunctionDef  # type: Union[Type[Any], Tuple[Type[Any], Type[Any]]]  # noqa: E501
if hasattr(ast, 'AsyncFunctionDef'):
    FunctionDef = (ast.FunctionDef, ast.AsyncFunctionDef)


def read_program(filename):  # type: (str) -> Union[bytes, str]
    """Read a program from a file.

    Args:
        filename: The name of the file to read. If set to '-', then we will
            read from stdin.

    Returns:
        The program as a single string.

    """
    program = None  # type: Union[bytes, Optional[str]]
    if filename == '-':
        program = sys.stdin.read()
    else:
        with open(filename, 'rb') as fin:
            program = fin.read()
    return program or ''


def _get_arguments(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Tuple[List[str], List[str]]  # noqa: E501
    arguments = list()  # type: List[str]
    types = list()  # type: List[str]

    def add_arg_by_name(name, arg):
        arguments.append(name)
        if arg.annotation is not None and hasattr(arg.annotation, 'id'):
            types.append(arg.annotation.id)
        else:
            types.append(None)

    for arg in fn.args.args:
        add_arg_by_name(arg.arg, arg)

    for arg in fn.args.kwonlyargs:
        add_arg_by_name(arg.arg, arg)

    # Handle single-star arguments.
    if fn.args.vararg is not None:
        name = '*' + fn.args.vararg.arg
        add_arg_by_name(name, fn.args.vararg)

    if fn.args.kwarg is not None:
        name = '**' + fn.args.kwarg.arg
        add_arg_by_name(name, fn.args.kwarg)

    return arguments, types


def _walk(fun, skip):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Callable) -> Iterator[ast.AST]  # noqa: E501
    """Walk through the nodes in this function, skipping as necessary.

    ast.walk goes through nodes in an arbitrary order, and doesn't
    allow you to skip an entire branch (as far as I know.)  This
    function will perform an in-order breadth-first traversal, and will
    skip unnecessary branches.

    Args:
        fun: The function to walk.
        skip: A function which returns True if we should skip the current
            node and all of its children.

    Yields:
        Children of the function and the function itself.

    """
    queue = deque()  # type: deque
    queue.appendleft(fun)
    while len(queue) > 0:
        curr = queue.pop()
        if skip(curr):
            continue
        if hasattr(curr, 'body'):
            queue.extendleft(curr.body)
        if hasattr(curr, 'handlers'):
            queue.extendleft(curr.handlers)
        if hasattr(curr, 'orelse'):
            queue.extendleft(curr.orelse)
        yield curr


def _has_return(fun):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool
    """Return true if the function has a fruitful return.

    Args:
        fun: A function node to check.

    Returns:
        True if there is a fruitful return, otherwise False.

    """
    def skip(f):
        return f != fun and isinstance(f, FunctionDef)

    for node in _walk(fun, skip):
        if isinstance(node, ast.Return) and node.value is not None:
            return True
    return False


def _has_yield(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool  # noqa: E501
    for node in ast.walk(fun):
        if isinstance(node, ast.Yield) or isinstance(node, ast.YieldFrom):
            return True
    return False


def _get_docstring(fun):  # type: (ast.AST) -> str
    return ast.get_docstring(fun)


def _get_all_functions(tree):  # type: (ast.AST) -> Iterator[Union[ast.FunctionDef, ast.AsyncFunctionDef]]  # noqa: E501
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            yield node
        elif hasattr(ast, 'AsyncFunctionDef'):
            if isinstance(node, ast.AsyncFunctionDef):
                yield node


def _get_all_classes(tree):  # type: (ast.AST) -> Iterator[ast.ClassDef]
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            yield node


def _get_all_methods(tree):  # type: (ast.AST) -> Iterator[Union[ast.FunctionDef, ast.AsyncFunctionDef]]  # noqa: E501
    for klass in _get_all_classes(tree):
        for fun in _get_all_functions(klass):
            yield fun


def _get_decorator_names(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> List[str]  # noqa: E501
    """Get decorator names from the function.

    Args:
        fun: The function whose decorators we are getting.

    Returns:
        The names of the decorators. Does not include setters and
        getters.

    """
    ret = list()
    for decorator in fun.decorator_list:
        # Attributes (setters and getters) won't have an id.
        if hasattr(decorator, 'id'):
            ret.append(getattr(decorator, 'id'))
    return ret


def _is_classmethod(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool  # noqa: E501
    return 'classmethod' in _get_decorator_names(fun)


def _is_staticmethod(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool  # noqa: E501
    return 'staticmethod' in _get_decorator_names(fun)


def _get_stripped_method_args(method):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Tuple[List[str], List[str]]  # noqa: E501
    args, types = _get_arguments(method)
    if 'cls' in args and _is_classmethod(method):
        args.remove('cls')
        types.pop(0)
    elif 'self' in args and not _is_staticmethod(method):
        args.remove('self')
        types.pop(0)
    return args, types


def _get_all_raises(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Iterator[ast.Raise]  # noqa: E501
    for node in ast.walk(fn):
        if isinstance(node, ast.Raise):
            yield node


def _get_exception_name(raises):  # type: (ast.Raise) -> str
    if isinstance(raises.exc, ast.Name):
        return raises.exc.id
    elif isinstance(raises.exc, ast.Call):
        if hasattr(raises.exc.func, 'id'):
            return getattr(raises.exc.func, 'id')
        elif hasattr(raises.exc.func, 'attr'):
            return getattr(raises.exc.func, 'attr')
        else:
            logger.debug(
                'Raises function call has neither id nor attr.'
                'has only: %s' % str(dir(raises.exc.func))
            )
    elif isinstance(raises.exc, ast.Attribute):
        return raises.exc.attr
    elif isinstance(raises.exc, ast.Subscript):
        id_repr = ''
        if hasattr(raises.exc.value, 'id'):
            id_repr = getattr(raises.exc.value, 'id')
        n_repr = ''
        if hasattr(raises.exc.slice, 'value'):
            value = getattr(raises.exc.slice, 'value')
            if hasattr(value, 'n'):
                n_repr = getattr(value, 'n')
        return '{}[{}]'.format(
            id_repr,
            n_repr,
        )
    else:
        logger.debug('Unexpected type in raises expression: {}'.format(
            raises.exc
        ))
    return ''


def _get_exceptions_raised(fn):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Set[str]  # noqa: E501
    visitor = RaiseVisitor()
    visitor.visit(fn)
    return visitor.exceptions


def _get_return_type(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Optional[str]
    if fn.returns is not None and hasattr(fn.returns, 'id'):
        return getattr(fn.returns, 'id')
    return None


def _get_all_variable_names(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Iterator[str]
    """Get all variable names used in the function.

    Args:
        fn: The function.

    Yields:
        The nodes which represent names.

    """
    for node in ast.walk(fn):
        if isinstance(node, ast.Name):
            yield node.id


def get_line_number_from_function(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> int
    """Get the line number for the end of the function signature.

    The function signature can be farther down when the parameter
    list is split across multiple lines.

    Args:
        fn: The function from which we are getting the line number.

    Returns:
        The line number for the start of the docstring for this
        function.

    """
    line_number = fn.lineno
    if hasattr(fn, 'args') and fn.args.args:
        last_arg = fn.args.args[-1]
        line_number = last_arg.lineno
    return line_number


class FunctionDescription(object):
    """Describes a function or method.

    Whereas a `Docstring` object describes a function's docstring,
    a `FunctionDescription` describes the function itself.  (What,
    ideally, the docstring should describe.)

    """

    def __init__(self, is_method, function):
        # type: (bool, Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> None
        """Create a new FunctionDescription.

        Args:
            is_method: True if this is a method. Will attempt to remove
                self or cls if appropriate.
            function: The base node of the function.

        """
        self.is_method = is_method
        self.function = function
        self.line_number = get_line_number_from_function(function)
        self.name = function.name
        if is_method:
            self.argument_names, self.argument_types = (
                _get_stripped_method_args(function)
            )
        else:
            self.argument_names, self.argument_types = _get_arguments(function)
        self.has_return = _has_return(function)
        self.return_type = _get_return_type(function)
        self.has_yield = _has_yield(function)
        self.docstring = _get_docstring(function)
        try:
            self.raises = _get_exceptions_raised(function)
        except Exception as ex:
            msg = '{}: {}'.format(self.name, ex)
            logger.debug(msg)
        self.variables = _get_all_variable_names(function)


def get_function_descriptions(program):
    # type: (ast.AST) -> List[FunctionDescription]
    """Get function name, args, return presence and docstrings.

    This function should be called on the top level of the
    document (for functions), and on classes (for methods.)

    Args:
        program: The tree representing the entire program.

    Returns:
        A list of function descriptions pulled from the ast.

    """
    ret = list()  # type: List[FunctionDescription]

    methods = set(_get_all_methods(program))
    for method in methods:
        ret.append(FunctionDescription(is_method=True, function=method))

    functions = set(_get_all_functions(program)) - methods
    for function in functions:
        ret.append(FunctionDescription(is_method=False, function=function))

    return ret



from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)
from collections import (
    defaultdict,
)

from ..parse.identifiers import (
    ArgumentItemIdentifier,
    ArgumentTypeIdentifier,
    ExceptionItemIdentifier,
    ReturnTypeIdentifier,
    YieldTypeIdentifier,
    Identifier,
    NoqaIdentifier,
)
from .base import (
    BaseDocstring,
    DocstringStyle,
    Sections,
)
from ..node import (
    CykNode,
)
from ..config import (
    Configuration,
    Strictness,
)
from ..parse.numpy import (
    parse,
)
from ..lex import (
    lex,
    condense,
)
from ..errors import (
    DarglintError,
)


class Docstring(BaseDocstring):

    def __init__(self, root, style=DocstringStyle.SPHINX, config=None):
        # type: (Union[CykNode, str], DocstringStyle, Optional[Configuration]) -> None  # noqa: E501
        """Create a new docstring from the AST.

        Args:
            root: The root of the AST, or the docstring
                (as a string.)  If it is a string, the
                string will be parsed.
            style: The docstring style.  Discarded, since this
                docstring always represents the Numpy style.
            config: The configuration parameters.  Passed to
                lex to change the number of spaces which count
                as an indentation.

        """
        if isinstance(root, CykNode):
            self.root = root  # type: Optional[CykNode]
        else:
            self.root = parse(condense(lex(root, config)))
        self._lookup = self._discover()

    def _discover(self):
        # type: () -> Dict[str, List[CykNode]]
        """Walk the tree, finding all non-terminal nodes.

        Returns:
            A lookup table for compound Nodes by their NodeType.

        """
        if not self.root:
            return dict()
        lookup = defaultdict(
            lambda: list()
        )  # type: Dict[str, List[CykNode]]
        for node in self.root.in_order_traverse():
            lookup[node.symbol].append(node)
            for annotation in node.annotations:
                if issubclass(annotation, Identifier):
                    lookup[annotation.key].append(node)
        return lookup

    def get_section(self, section):
        # type: (Sections) -> Optional[str]
        nodes = []  # type: Optional[List[CykNode]]

        # TODO: Add Receives section
        if section == Sections.SHORT_DESCRIPTION:
            nodes = self._lookup.get('short-description', None)
        elif section == Sections.LONG_DESCRIPTION:
            nodes = self._lookup.get('long-description', None)
        elif section == Sections.ARGUMENTS_SECTION:
            nodes = self._lookup.get('arguments-section', None)
            extra = self._lookup.get('other-arguments-section', None)
            if nodes:
                nodes.extend(extra or [])
            else:
                nodes = extra
        elif section == Sections.RAISES_SECTION:
            nodes = self._lookup.get('raises-section', None)
            extra = self._lookup.get('warns-section', None)
            if nodes:
                nodes.extend(extra or [])
            else:
                nodes = extra
        elif section == Sections.YIELDS_SECTION:
            nodes = self._lookup.get('yields-section', None)
        elif section == Sections.RETURNS_SECTION:
            nodes = self._lookup.get('returns-section', None)
        elif section == Sections.NOQAS:
            nodes = self._lookup.get('noqa', None)
        else:
            raise Exception(
                'Unsupported section type, {}'.format(section)
            )

        if not nodes:
            return None

        return_value = ''
        for node in nodes:
            return_value += '\n\n' + node.reconstruct_string()

        return return_value.strip() or None

    def _get_types_unsorted(self, section):
        # type: (Sections) -> Optional[Union[str, List[Optional[str]]]]
        if section == Sections.ARGUMENTS_SECTION:
            if 'arguments-section' not in self._lookup:
                return None
            return [
                ArgumentTypeIdentifier.extract(x)
                for x in self._lookup.get(ArgumentTypeIdentifier.key, [])
            ]
        else:
            raise Exception(
                'Section type {} does not have types, '.format(
                    section.name
                ) + 'or is not yet supported'
            )
        return None

    def get_types(self, section):
        # type: (Sections) -> Optional[Union[str, List[Optional[str]]]]
        if section == Sections.RETURNS_SECTION:
            return_type = self._lookup.get(ReturnTypeIdentifier.key, [])
            if len(return_type) == 0:
                return None
            elif len(return_type) == 1:
                return ReturnTypeIdentifier.extract(return_type[0])
            else:
                raise NotImplementedError(
                    'Multiple types should be combined into a Union'
                )
        elif section == Sections.YIELDS_SECTION:
            yield_type = self._lookup.get(YieldTypeIdentifier.key, [])
            if len(yield_type) == 0:
                return None
            elif len(yield_type) == 1:
                return YieldTypeIdentifier.extract(yield_type[0])
            else:
                raise NotImplementedError(
                    'Multiple types should be combined into a Union'
                )

        # Sort the types according to the sorted items section.
        # This ensures that the results are always in the same
        # order.
        items = self._get_items_unsorted(section)
        if not items:
            return None
        types = self._get_types_unsorted(section)
        if not types:
            return []

        # Handle the case when two arguments are listed in
        # one line (like "m, n: int").
        sorted_types = list()
        for arg, _type in sorted(zip(items, types)):
            for _ in arg.split(','):
                sorted_types.append(_type)
        return sorted_types

    def _get_items_unsorted(self, section):
        # type: (Sections) -> Optional[List[str]]
        if section == Sections.ARGUMENTS_SECTION:
            return [
                ArgumentItemIdentifier.extract(x)
                for x in self._lookup.get(ArgumentItemIdentifier.key, [])
            ] or None
        elif section == Sections.RAISES_SECTION:
            return sorted(
                ExceptionItemIdentifier.extract(x)
                for x in self._lookup.get(ExceptionItemIdentifier.key, [])
            ) or None
        else:
            raise Exception(
                'Section type {} does not have items, '.format(
                    section.name
                ) + 'or is not yet supported.'
            )
        return None

    def get_items(self, section):
        # type: (Sections) -> Optional[List[str]]
        items = self._get_items_unsorted(section)
        if not items:
            return None
        sorted_items = list()
        for item in sorted(items):
            sorted_items.extend([
                x.strip() for x in item.split(',')
            ])
        return sorted_items

    def get_noqas(self):
        # type: () -> Dict[str, List[str]]
        """Get a map of the errors ignored to their targets.

        Returns:
            A dictionary containing the errors to ignore as keys and
            a list of which targets to apply these exceptions to as
            the values.  A blank list implies a global noqa.

        """
        noqas = dict()
        for noqa in self._lookup[NoqaIdentifier.key]:
            noqas[NoqaIdentifier.extract(noqa) or '*'] = (
                NoqaIdentifier.extract_targets(noqa)
            )
        return noqas

    def get_line_numbers(self, node_type):
        # type: (str) -> Optional[Tuple[int, int]]
        """Get the line numbers for the first instance of the given section.

        Args:
            node_type: The NodeType which we want line numbers for.
                These should be unique instances. (I.e. they should be
                in the set of compound NodeTypes which only occur
                once in a docstring. For example, "Raises" and "Args".

        Returns:
            The line numbers for the first instance of the given node type.

        """
        nodes = self._lookup[node_type]
        if nodes:
            return nodes[0].line_numbers
        return None

    def get_line_numbers_for_value(self, node_type, value):
        # type: (str, str) -> Optional[Tuple[int, int]]
        """Get the line number for a node with the given value.

        Args:
            node_type: The compound node which should contain the
                node we are searching for.
            value: The value of the node.

        Returns:
            A list of line numbers for nodes which match the
            parameters.

        """
        nodes = self._lookup[node_type]
        for node in nodes:
            for child in node.walk():
                if child.value == value and child.line_numbers:
                    return child.line_numbers
        return None

    @property
    def ignore_all(self):
        # type: () -> bool
        """Return whether we should ignore everything in the docstring.

        This happens when there is a bare noqa in the docstring, or
        there is "noqa: *" in the docstring.

        Returns:
            True if we should ignore everything, otherwise false.

        """
        return False

    def get_style_errors(self):
        # type: () -> Iterable[Tuple[Callable, Tuple[int, int]]]
        """Get any style errors annotated on the tree.

        Yields:
            Instances of DarglintErrors for style issues.

        # noqa: I302

        """
        if not self.root:
            return
        for node in self.root.in_order_traverse():
            for annotation in node.annotations:
                if issubclass(annotation, DarglintError):
                    yield annotation, node.line_numbers

    def satisfies_strictness(self, strictness):
        # type: (Strictness) -> bool
        """Return true if the docstring has no more than the min strictness.

        Args:
            strictness: The minimum amount of strictness which should
                be present in the docstring.

        Returns:
            True if there is no more than the minimum amount of strictness.

        """
        sections = {
            section for section in Sections
            if self.get_section(section)
        }
        if strictness == Strictness.SHORT_DESCRIPTION:
            return sections == {Sections.SHORT_DESCRIPTION}
        elif strictness == Strictness.LONG_DESCRIPTION:
            return sections in [
                {Sections.SHORT_DESCRIPTION},
                # Shouldn't be possible, but if it is in the future, then
                # we should allow this.
                {Sections.LONG_DESCRIPTION},
                {Sections.SHORT_DESCRIPTION, Sections.LONG_DESCRIPTION},
            ]
        else:
            return False



"""This module reads the configuration file."""

import configparser
from enum import Enum
import logging
from logging import (  # noqa
    Logger,
)
import os

from typing import (  # noqa
    Iterable,
    List,
    Optional,
)

from .docstring.base import DocstringStyle


# TODO: Configure this logger and allow the user to specify
# the whether they want warnings.
logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger('darglint')

POSSIBLE_CONFIG_FILENAMES = (
    '.darglint',
    'setup.cfg',
    'tox.ini',
)

DEFAULT_DISABLED = {'DAR104'}


class Strictness(Enum):
    """The minimum strictness with which to apply checks.

    Strictness does not describe whether or not a check
    should be applied. Rather, if a check is done, strictness
    describes how intense/strict/deep the check should be.

    Each level here describes what is required of the
    docstring at the given level of strictness.  For example,
    SHORT_DESCRIPTION describes the situation where one-liners are
    allowed, and sections are not required.

    If the docstring being checked contains more than the
    allowed amount below, then it is assumed that everything
    must be checked.

    """

    # Allow a single-line description.
    SHORT_DESCRIPTION = 1

    # Allow a single-line description followed by a long
    # description, but no sections.
    LONG_DESCRIPTION = 2

    # Require everything.
    FULL_DESCRIPTION = 3


class Configuration(object):

    def __init__(self, ignore, message_template, style, strictness,
                 ignore_regex=None, enable=[], indentation=4):
        # type: (List[str], Optional[str], DocstringStyle, Strictness, Optional[str], List[str], int) -> None  # noqa: E501
        """Initialize the configuration object.

        Args:
            ignore: A list of error codes to ignore.
            message_template: the template with which to format the errors.
            style: The style of docstring.
            strictness: The minimum strictness to allow.
            ignore_regex: A regular expression which enables ignoring
                functions/methods by name.
            enable: A list of of error codes that are disabled by default.
            indentation: The number of spaces to count as an indent.

        """
        self._enable = enable
        self._ignore = ignore
        self.message_template = message_template
        self.style = style
        self.strictness = strictness
        self.errors_to_ignore = self._get_errors_to_ignore()
        self.ignore_regex = ignore_regex
        self.indentation = indentation

    @property
    def enable(self):
        # type: () -> List[str]
        return self._enable

    @enable.setter
    def enable(self, errors):
        # type: (List[str]) -> None
        self._enable = errors
        self.errors_to_ignore = self._get_errors_to_ignore()

    @property
    def ignore(self):
        # type: () -> List[str]
        return self._ignore

    @ignore.setter
    def ignore(self, errors):
        # type: (List[str]) -> None
        self._ignore = errors
        self.errors_to_ignore = self._get_errors_to_ignore()

    def _get_errors_to_ignore(self):
        # type: () -> List[str]
        """Update the errors to ignore, accounding for defaults.

        For use in constructing a cached `errors_to_ignore` value.
        Since this value could be used frequently, it makes
        sense to cache this value.

        Returns:
            The errors to ignore, including the default errors.

        """
        disabled = DEFAULT_DISABLED - set(self._enable)
        return self._ignore + list(disabled)


def load_config_file(filename):  # type: (str) -> Configuration
    """Load the config file located at the filename.

    Args:
        filename: A valid filename to read from.

    Raises:
        Exception: When the configuration style is not a valid choice.

    Returns:
        A Configuration object.

    """
    config = configparser.ConfigParser()
    config.read(filename)
    ignore = list()
    enable = list()
    message_template = None
    ignore_regex = None
    style = DocstringStyle.GOOGLE
    strictness = Strictness.FULL_DESCRIPTION
    indentation = 4
    if 'darglint' in config.sections():
        if 'ignore' in config['darglint']:
            errors = config['darglint']['ignore']
            for error in errors.split(','):
                ignore.append(error.strip())
        if 'enable' in config.sections():
            to_enable = config['darglint']['enable']
            for error in to_enable.split(','):
                enable.append(error.strip())
        if 'message_template' in config['darglint']:
            message_template = config['darglint']['message_template']
        if 'ignore_regex' in config['darglint']:
            ignore_regex = config['darglint']['ignore_regex']
        if 'docstring_style' in config['darglint']:
            raw_style = config['darglint']['docstring_style'].lower().strip()
            if raw_style == 'google':
                style = DocstringStyle.GOOGLE
            elif raw_style == 'sphinx':
                style = DocstringStyle.SPHINX
            elif raw_style == 'numpy':
                style = DocstringStyle.NUMPY
            else:
                raise Exception(
                    'Unrecognized style.  Should be one of {}'.format(
                        [x.name for x in DocstringStyle]
                    )
                )
        if 'strictness' in config['darglint']:
            raw_strictness = config['darglint']['strictness'].lower().strip()
            if raw_strictness in {'short_description', 'short'}:
                strictness = Strictness.SHORT_DESCRIPTION
            elif raw_strictness in {'long_description', 'long'}:
                strictness = Strictness.LONG_DESCRIPTION
            elif raw_strictness in {'full_description', 'full'}:
                strictness = Strictness.FULL_DESCRIPTION
            else:
                raise Exception(
                    'Unrecognized stricteness amount.  '
                    'Should be one of {"short", "long", "full"}'
                )
        if 'indentation' in config['darglint']:
            try:
                indentation = int(config['darglint']['indentation'])
            except ValueError:
                raise Exception(
                    'Unrecognized value for indentation.  Expected '
                    'a non-zero, positive integer, but received {}'.format(
                        config['darglint']['indentation']
                    )
                )
    return Configuration(
        ignore=ignore,
        message_template=message_template,
        style=style,
        strictness=strictness,
        ignore_regex=ignore_regex,
        enable=enable,
        indentation=indentation,
    )


def walk_path():  # type: () -> Iterable[str]
    """Yield directories from the current to root.

    Yields:
        The current directory, then its parent, etc. all
        the way up to root.

    """
    cwd = os.getcwd()
    yield cwd
    prev = cwd
    next_path = os.path.dirname(cwd)

    # Assumes that os.path.dirname will give the root path back
    # when given the root path.
    while prev != next_path:
        yield next_path
        prev = next_path
        next_path = os.path.dirname(next_path)


def find_config_file_in_path(path):  # type: (str) -> Optional[str]
    """Return the config path, if it is correct, or None.

    Args:
        path: The path to check.

    Returns:
        The fully qualified path to the config file, if it is
        in this directory, otherwise none.

    """
    filenames = os.listdir(path)
    for filename in filenames:
        if filename in POSSIBLE_CONFIG_FILENAMES:
            config = configparser.ConfigParser()
            fully_qualified_path = os.path.join(path, filename)
            try:
                config.read(fully_qualified_path)
                if 'darglint' in config.sections():
                    return fully_qualified_path
            except configparser.ParsingError:
                logger.error('Unable to parse file {}'.format(
                    fully_qualified_path
                ))
    return None


def find_config_file():  # type: () -> Optional[str]
    """Return the location of the config file.

    Returns:
        The location of the config file, if it exists.
        Otherwise, returns None.

    """
    # Check the current directory
    for path in walk_path():
        possible_config_filename = find_config_file_in_path(path)
        if possible_config_filename is not None:
            return possible_config_filename
    return None


def get_config():  # type: () -> Configuration
    """Locate the configuration file and return its Configuration.

    Returns:
        The Configuration described in the nearest configuration file,
        otherwise an empty Configuration.

    """
    filename = find_config_file()
    if filename is None:
        return Configuration(
            ignore=list(),
            message_template=None,
            style=DocstringStyle.GOOGLE,
            strictness=Strictness.FULL_DESCRIPTION,
        )
    return load_config_file(filename)


def get_logger():  # type: () -> Logger
    """Get the default logger for darglint.

    Returns:
        The default logger for darglint.

    """
    return logger



from typing import (
    List,
    Optional,
)
from functools import (
    reduce,
)
from ..token import (
    Token,
    TokenType,
)
from ..node import (
    CykNode,
)
from ..peaker import (
    Peaker,
)
from .identifiers import (
    NoqaIdentifier,
)


def _is(peaker, token_type, index=1):
    # type: (Peaker[Token], Optional[TokenType], int) -> bool
    try:
        token = peaker.peak(lookahead=index)
    except IndexError:
        token = None
    if not token_type and not token:
        return True
    return bool(token and token.token_type == token_type)


def _are(peaker, *token_types):
    # type: (Peaker[Token], Optional[TokenType]) -> bool
    return all([
        _is(peaker, token_type, i + 1)
        for i, token_type in enumerate(token_types)
    ])


def _parse_noqa_head(peaker):
    # type: (Peaker[Token]) -> Optional[CykNode]
    if not (
        _are(peaker, TokenType.HASH, TokenType.NOQA, TokenType.NEWLINE)
        or _are(peaker, TokenType.HASH, TokenType.NOQA, None)
    ):
        return None
    noqa_hash = CykNode('hash', value=peaker.next())
    noqa = CykNode('noqa', value=peaker.next())
    if _is(peaker, TokenType.NEWLINE):
        peaker.next()
    return CykNode(
        'noqa',
        lchild=noqa_hash,
        rchild=noqa,
        annotations=[
            NoqaIdentifier,
        ],
    )


def _last_node(node):
    # type: (CykNode) -> CykNode
    curr = node
    rchild = curr.rchild
    while rchild:
        curr = rchild
        rchild = curr.rchild
    return curr


def foldr(fun, xs, acc):
    return reduce(lambda x, y: fun(y, x), xs[::-1], acc)


def _parse_words_until_newline_or_end(peaker):
    if not peaker.has_next() or _is(peaker, TokenType.NEWLINE):
        return None
    words = [CykNode('word', value=peaker.next())]
    while peaker.has_next() and not _is(peaker, TokenType.NEWLINE):
        words.append(CykNode('word', value=peaker.next()))

    if len(words) == 1:
        head = words[0]
        head.symbol = 'words'
        return head

    def join(x, y):
        return CykNode(
            'words',
            lchild=x,
            rchild=y,
        )

    acc = words.pop()
    acc.symbol = 'words'

    return foldr(join, words, acc)


def _parse_noqa(peaker):
    # type: (Peaker[Token]) -> Optional[CykNode]
    if not (
        _are(peaker, TokenType.HASH, TokenType.NOQA, TokenType.COLON,
             TokenType.WORD)
    ):
        return None
    noqa_hash = CykNode('hash', value=peaker.next())
    noqa = CykNode('noqa', value=peaker.next())
    colon = CykNode('colon', value=peaker.next())
    targets = _parse_words_until_newline_or_end(peaker)
    head = CykNode(
        'noqa',
        lchild=CykNode(
            'noqa-head',
            lchild=noqa_hash,
            rchild=noqa,
        ),
        rchild=CykNode(
            'noqa-statement1',
            lchild=colon,
            rchild=targets,
        ),
        annotations=[
            NoqaIdentifier,
        ],
    )
    return head


def _parse_long_description(peaker):
    # type: (Peaker[Token]) -> Optional[CykNode]
    if not peaker.has_next():
        return None
    head = _parse_noqa(peaker) or _parse_noqa_head(peaker)
    if head:
        new_head = CykNode(
            'long-description',
            lchild=head,
        )
        head = new_head
    else:
        head = CykNode(
            symbol='long-description',
            lchild=CykNode('long-description1', value=peaker.next()),
        )
    curr = _last_node(head)
    while peaker.has_next():
        noqa = _parse_noqa(peaker) or _parse_noqa_head(peaker)
        if not noqa:  # curr.rchild:
            curr.rchild = CykNode(
                symbol='long-description1',
                lchild=CykNode('long-description1', value=peaker.next()),
            )
        else:
            old_left = curr.lchild
            curr.lchild = CykNode(
                symbol='long-description1',
                lchild=old_left,
                rchild=noqa,
            )
        curr = _last_node(curr)
    return head


def parse(tokens):
    # type: (List[Token]) -> Optional[CykNode]
    peaker = Peaker((x for x in tokens), lookahead=5)
    if not peaker.has_next():
        return None

    return _parse_long_description(peaker)



"""Defines the tokens that result from lexing, as well as their types."""

from enum import Enum


class BaseTokenType(Enum):
    pass


class TokenType(BaseTokenType):
    """The possible types of tokens."""

    WORD = 1
    COLON = 2
    DOCTERM = 3  # """
    NEWLINE = 4
    INDENT = 5  # Assumed to always be 4 spaces
    HASH = 6  # A hash mark (for comments/noqa).
    LPAREN = 7  # (
    RPAREN = 8  # )

    RETURNS = 9
    YIELDS = 10
    RAISES = 11
    ARGUMENTS = 12
    NOQA = 13
    RETURN_TYPE = 14
    YIELD_TYPE = 15
    VARIABLES = 16
    VARIABLE_TYPE = 17
    ARGUMENT_TYPE = 18
    HEADER = 19
    OTHER = 20
    RECEIVES = 21
    WARNS = 22
    SEE = 23
    ALSO = 24
    NOTES = 25
    EXAMPLES = 26
    REFERENCES = 27

    # next: 28


KEYWORDS = {
    TokenType.RETURNS,
    TokenType.YIELDS,
    TokenType.RAISES,
    TokenType.ARGUMENTS,
    TokenType.NOQA,
    TokenType.RETURN_TYPE,
    TokenType.YIELD_TYPE,
    TokenType.VARIABLES,
    TokenType.VARIABLE_TYPE,
    TokenType.ARGUMENT_TYPE,
    TokenType.RECEIVES,
    TokenType.WARNS,
    TokenType.NOTES,
    TokenType.EXAMPLES,
    TokenType.REFERENCES,
}


class Token(object):
    """A token representing anything which can appear in a docstring."""

    def __init__(self,
                 value: str,
                 token_type: TokenType,
                 line_number: int
                 ) -> None:
        """Create a new Token.

        Args:
            value: The value of the token. (The actual string.)
            token_type: The type of token this represents.
            line_number: The line number where this token resides.
                Used when reporting errors.

        """
        self.value = value
        self.token_type = token_type
        self.line_number = line_number

    def __str__(self):
        """Return readable representation for debugging.

        Returns:
            A readable representation for debugging.

        """
        return '<Token {} {}>'.format(repr(self.value), self.token_type)

    def __repr__(self):
        """Return readable representation for debugging.

        Returns:
            A readable representation for debugging.

        """
        return str(self)



import inspect
from typing import (
    List,
)
from functools import (
    reduce,
)

from ..token import (
    Token,
    TokenType,
    KEYWORDS,
)
from .cyk import (
    parse as cyk_parse,
)
from ..node import (
    CykNode,
)
from .combinator import (
    parser_combinator,
)
from .long_description import (
    parse as long_description_parse,
)

from .grammars.sphinx_arguments_section import ArgumentsGrammar
from .grammars.sphinx_argument_type_section import ArgumentTypeGrammar
from .grammars.sphinx_variables_section import VariablesSectionGrammar
from .grammars.sphinx_variable_type_section import VariableTypeGrammar
from .grammars.sphinx_raises_section import RaisesGrammar
from .grammars.sphinx_returns_section import ReturnsGrammar
from .grammars.sphinx_return_type_section import ReturnTypeGrammar
from .grammars.sphinx_short_description import ShortDescriptionGrammar
from .grammars.sphinx_yields_section import YieldsGrammar
from .grammars.sphinx_yield_type_section import YieldTypeGrammar


def two_newline_separated_or_keyword(tokens, i):
    # type: (List[Token], int) -> int
    newline_count = 0
    j = i
    while j < len(tokens):
        if tokens[j].token_type == TokenType.NEWLINE:
            newline_count += 1
        else:
            break
        j += 1

    if newline_count >= 2:
        return j

    if (j + 1 < len(tokens)
            and tokens[j].token_type == TokenType.COLON
            and tokens[j + 1].token_type in KEYWORDS):
        return j

    return 0


def top_parse(tokens):
    # type: (List[Token]) -> List[List[Token]]
    all_sections = list()
    curr = 0
    # Strip leading newlines.
    while curr < len(tokens) and tokens[curr].token_type == TokenType.NEWLINE:
        curr += 1
    prev = curr

    while curr < len(tokens):
        split_end = two_newline_separated_or_keyword(tokens, curr)
        if split_end > curr:
            if tokens[prev:curr]:
                all_sections.append(
                    tokens[prev:curr]
                )
            curr = split_end
            prev = curr
        else:
            curr += 1
    last_section = tokens[prev:curr]
    if last_section:
        all_sections.append(last_section)
    return all_sections


def _match(token):
    """Match the given token from the given section to a set of grammars.

    Args:
        token: The token to match.  This should hint at what sections
            could possibly be here.

    Returns:
        A list of grammars to be tried in order.

    """
    tt_lookup = {
        TokenType.VARIABLES: [
            VariablesSectionGrammar,
            long_description_parse,
        ],
        TokenType.ARGUMENTS: [
            ArgumentsGrammar,
            long_description_parse,
        ],
        TokenType.ARGUMENT_TYPE: [
            ArgumentTypeGrammar,
            long_description_parse,
        ],
        TokenType.VARIABLE_TYPE: [
            VariableTypeGrammar,
            long_description_parse,
        ],
        TokenType.RAISES: [
            RaisesGrammar,
            long_description_parse,
        ],
        TokenType.YIELDS: [
            YieldsGrammar,
            long_description_parse,
        ],
        TokenType.YIELD_TYPE: [
            YieldTypeGrammar,
            long_description_parse,
        ],
        TokenType.RETURNS: [
            ReturnsGrammar,
            long_description_parse,
        ],
        TokenType.RETURN_TYPE: [
            ReturnTypeGrammar,
            long_description_parse,
        ],
    }
    return tt_lookup.get(token.token_type, [long_description_parse])


def lookup(section, section_index=-1):
    assert len(section) > 0
    if (section[0].token_type == TokenType.COLON
            and len(section) > 1):
        grammars = _match(section[1])
    else:
        grammars = [long_description_parse]
    if section_index == 0:
        # Add the short description right before the long
        # description.  That way, the short description will
        # always be captured, if possible, and any other
        # possible sections will be captured, if possible.
        #
        # Assumes that the long description will always come
        # after all other possible grammars, if it appears at
        # all.
        if long_description_parse not in grammars:
            grammars.append(ShortDescriptionGrammar)
            return grammars
        long_index = grammars.index(long_description_parse)
        grammars.insert(long_index, ShortDescriptionGrammar)
        return grammars
    return grammars


def combinator(*args):
    def inner(*nodes):
        if len(nodes) == 1:
            return CykNode(
                symbol='docstring',
                lchild=nodes[0],
            )
        elif len(nodes) == 2:
            return CykNode(
                symbol='docstring',
                lchild=nodes[0],
                rchild=nodes[1],
            )
    if args:
        return reduce(inner, args)
    else:
        # The arguments are empty, so we return an
        # empty docstring.
        return CykNode(symbol='docstring')


def parse(tokens):
    def mapped_lookup(section, section_index=-1):
        for grammar in lookup(section, section_index):
            if inspect.isclass(grammar):
                yield lambda x: cyk_parse(grammar, x)
            else:
                yield grammar
    return parser_combinator(top_parse, mapped_lookup, combinator, tokens)



"""The error reporting classes."""

import ast  # noqa
from collections import OrderedDict
from typing import (  # noqa
    Dict,
    Iterator,
    List,
    Tuple,
    Union,
)
from .function_description import (
    get_line_number_from_function,
)

from .errors import DarglintError  # noqa


class ErrorReport(object):
    """Reports the errors for the given run."""

    def __init__(
        self,
        errors,
        filename,
        verbosity=2,
        message_template=None,
    ):
        # type: (List[DarglintError], str, int, str) -> None
        """Create a new error report.

        Args:
            errors: A list of DarglintError instances.
            filename: The name of the file the error came from.
            verbosity: A number in the set, {1, 2}, representing low
                and high verbosity.
            message_template: A python format string for specifying
                how the string representation of this ErrorReport
                should appear.

        """
        self.filename = filename
        self.verbosity = verbosity
        self.errors = errors
        self.error_dict = self._group_errors_by_function()
        if message_template is None:
            self.message_template = '{path}:{obj}:{line}: {msg_id}: {msg}'
        else:
            self.message_template = message_template

    def _sort(self):
        # type: () -> None
        self.errors.sort(key=lambda x: x.function.lineno)

    def _group_errors_by_function(self):
        # type: () -> Dict[Union[ast.FunctionDef, ast.AsyncFunctionDef], List[DarglintError]]  # noqa: E501
        """Sort the current errors by function, and put into an OrderedDict.

        Returns:
            An ordered dictionary of functions and their errors.

        """
        self._sort()
        error_dict = OrderedDict()  # type: Dict
        current = None  # The current function
        for error in self.errors:
            if current != error.function:
                current = error.function
                error_dict[current] = list()
            error_dict[current].append(error)

        # Sort all of the errors returned by the function
        # alphabetically.
        for key in error_dict:
            error_dict[key].sort(key=lambda x: x.message() or '')

        # Sort all of the errors returned by the key
        # by the line numbers.
        for key in error_dict:
            error_dict[key].sort(key=lambda x: x.line_numbers or (0, 0))

        return error_dict

    def _get_error_description(self, error):  # type: (DarglintError) -> str
        """Get the error description.

        Args:
            error: The error to describe.

        Returns:
            A string representing the error.

        """
        line_number = get_line_number_from_function(error.function)
        if (hasattr(error.function, 'decorator_list')
                and error.function.decorator_list):
            line_number += len(error.function.decorator_list)
        if error.line_numbers:
            line_number += error.line_numbers[0] + 1
        return self.message_template.format(
            msg_id=error.error_code,
            msg=error.message(verbosity=self.verbosity),
            path=self.filename,
            obj=error.function.name,
            line=line_number,  # error.function.lineno,
        )

    def __str__(self):  # type: () -> str
        """Return a string representation of this error report.

        Returns:
            A string representation of this error report.

        """
        if len(self.errors) == 0:
            return ''
        ret = list()
        for function in self.error_dict:
            for error in self.error_dict[function]:
                ret.append(self._get_error_description(error))
        return '\n'.join(ret)

    def flake8_report(self):
        # type: () -> Iterator[Tuple[int, int, str]]
        # line, col, message
        for function in self.error_dict:
            for error in self.error_dict[function]:
                # TODO: Shouldn't get_line_number (here and above) return
                # the correct line number?  Why do we have to handle decorators
                # here?
                line_number = get_line_number_from_function(error.function)
                if (hasattr(error.function, 'decorator_list')
                        and error.function.decorator_list):
                    line_number += len(error.function.decorator_list)
                if error.line_numbers:
                    line_number += error.line_numbers[0] + 1
                else:
                    line_number += 1
                # TODO: Do we need verbosity here?
                message = '{} {}'.format(
                    error.error_code,
                    error.message(self.verbosity),
                )
                yield (line_number, 0, message)



r"""A parser combinator.

This combinator first parses a docstring into sections
(any block separated by two newlines), then calls individual
CYK parsers for each subsection.   The subparsers are called
in order of specificity.

Although weighting an overall CYK parser could handle an
entire docstring, this should be much more efficient.  A CYK
parser is O(n^3).  By reducing the docstring into k separate
sections, we can reduce this to

    O(n + k(n / k)^3)
    = O(n + k (n^3 / k^3))
    = O(n + (n^3 / k^2))

Which is still a member of O(n^3), since k is constant.  However,
since our k is around 4 members, this constant divisor will
actually impart a significant improvement, even though it doesn't
change the order.

Upon inspection, threading did not significantly improve performance.
The probem, it seems, is that some of the sections take much
longer than others, and so threading does nothing to improve
speed. (It actually made it worse.)

"""


def parser_combinator(top, lookup, combinator, tokens):
    """Parse the given tokens, combining in the given fashion.

    Args:
        top: The top-level parser.  Separates the tokens into
            sections which can be consumed by the parsers in the
            lookup function.
        lookup: For a given section from the top-level parser,
            returns a list of possible parsers.
        combinator: Combines the resultant nodes from parsing
            each section from the top-level parser.
        tokens: The tokens to be parsed.

    Returns:
        The top-level node from the combinator.

    """
    sections = top(tokens)
    parsed_sections = list()
    for i, section in enumerate(sections):
        parsed = None
        for parse in lookup(section, i):
            parsed = parse(section)
            if parsed:
                break
        if not parsed:
            return None
        parsed_sections.append(parsed)
    return combinator(*parsed_sections)



"""Describes Peaker, a stream transformer for peaking ahead."""
from collections import deque

from typing import (  # noqa: F401
    Callable,
    Generic,
    Iterator,
    List,
    TypeVar,
    Union,
    Optional,
)


T = TypeVar('T')


class Peaker(Generic[T]):
    """A stream transformer allowing us to peak ahead."""

    # The previous token which was gotten.
    prev = None  # type: T

    class _Empty(object):
        value = None

    def __init__(self, stream, lookahead=1):
        # type: (Iterator[T], int) -> None
        """Create a new peaker.

        Args:
            stream: An iterator of T objects, which may be empty.
            lookahead: The amount of lookahead this should allow
                in the stream.

        """
        self.stream = stream
        self.buffer = deque()  # type: deque
        self.lookahead = lookahead
        self._buffer_to(lookahead)

    def _buffer_to(self, amount):
        """Extend the internal buffer to the given amount.

        Only adds items up to that amount, and while there
        are items to add.

        Args:
            amount: The length to make the buffer.

        Raises:
            Exception: If we are not able to buffer to the
                given amount.

        """
        if amount > self.lookahead:
            raise Exception(
                'Cannot extend buffer to {}: '
                'beyond buffer lookahead {}'.format(
                    amount, self.lookahead
                )
            )
        while len(self.buffer) < amount:
            try:
                self.buffer.appendleft(next(self.stream))
            except StopIteration:
                break

    def next(self):
        # type: () -> T
        """Get the next item in the stream, moving it forward.

        Side effects:
            Moves the stream forward.

        Raises:
            StopIteration: If there are no more items in the stream.

        Returns:
            The next item of type T in the stream.

        """
        if len(self.buffer) == 0:
            raise StopIteration
        self.prev = self.buffer.pop()
        self._buffer_to(self.lookahead)
        return self.prev

    def peak(self, lookahead=1):
        # type: (int) -> Optional[T]
        """Get the next letter in the stream, without moving it forward.

        Args:
            lookahead: The amount of tokens to look ahead in
                the buffer.

        Raises:
            Exception: If we are not able to buffer to the given
                lookahead.

        Returns:
            The next item of type T in the stream.

        """
        if lookahead > self.lookahead:
            raise Exception(
                'Cannot peak to {}: beyond buffer lookahead {}'.format(
                    lookahead, self.lookahead
                )
            )
        if lookahead > len(self.buffer):
            return None
        index = len(self.buffer) - lookahead
        return self.buffer[index]

    def rpeak(self, lookahead=1):
        # type: (int) -> T
        """Peak at the item lookahead ahead, raising an exception if empty.

        Args:
            lookahead: The amount of tokens to look ahead in
                the buffer.

        Raises:
            Exception: If we are not able to buffer to the given
                lookahead.
            IndexError: If there are no items at the given index ahead.

        Returns:
            The next item of type T in the stream.

        """
        if lookahead > self.lookahead:
            raise Exception(
                'Cannot peak to {}: beyond buffer lookahead {}'.format(
                    lookahead, self.lookahead
                )
            )
        if lookahead > len(self.buffer):
            raise IndexError
        index = len(self.buffer) - lookahead
        return self.buffer[index]

    def has_next(self):
        # type: () -> bool
        """Tell whether there are more tokens in the stream.

        Returns:
            True if there are more tokens, false otherwise.

        """
        return len(self.buffer) > 0

    def take_while(self, test):
        # type: (Callable) -> List[T]
        """Return elements from the stream while they pass the test.

        Args:
            test: A function which returns true if we would like to collect
                the token, or false if we would like to stop.

        Returns:
            A list of items (of type T), which pass the given test function.

        """
        passing_elements = []  # type: List[T]
        while self.has_next() and test(self.peak()):
            passing_elements.append(self.next())
        return passing_elements



import abc

from .cyk import CykNode
from typing import (
    List,
)
from ..token import (
    TokenType,
)


class Identifier(abc.ABC):
    """Identifies a particular target in the grammar.

    Since the transformation from BNF to CNF messes up
    the grammars, we can use this class to mark pieces
    of the grammar we need access to.  This should greatly
    simplify the process of interpreting the produced tree.

    """

    @property
    @abc.abstractmethod
    def key(self):
        # type: () -> str
        pass

    @staticmethod
    @abc.abstractmethod
    def extract(node):
        # type: (CykNode) -> str
        pass

    def __str__(self):
        return '<identifier: {}>'.format(self.key)

    def __repr__(self):
        return str(self)


# TODO: Write a class to extract a given value from a base node
# given a path.  Use that to simplify the extract methods below.


class ArgumentItemIdentifier(Identifier):

    key = 'id_ArgsItem'

    @staticmethod
    def extract(node):
        return node.lchild.value.value


class ArgumentIdentifier(Identifier):

    key = 'id_Args'

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        assert node.rchild
        assert node.rchild.lchild
        assert node.rchild.lchild.value
        assert node.rchild.lchild.value.value
        return node.rchild.lchild.value.value


class ArgumentTypeIdentifier(Identifier):

    key = 'id_ArgType'

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        assert node.rchild
        assert node.rchild.rchild
        assert node.rchild.rchild.lchild
        if node.rchild.rchild.lchild.rchild:
            assert node.rchild.rchild.lchild.rchild.lchild
            if node.rchild.rchild.lchild.rchild.lchild.value:
                return node.rchild.rchild.lchild.rchild.lchild.value.value
            else:
                return (
                    node.rchild.rchild.lchild.rchild
                        .lchild.reconstruct_string()
                )
        else:
            if node.rchild.rchild.lchild.value:
                return node.rchild.rchild.lchild.value.value
            return node.rchild.rchild.lchild.reconstruct_string()


class ExceptionItemIdentifier(Identifier):

    key = 'id_ExceptItem'

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        assert node.lchild
        assert node.lchild.value
        return node.lchild.value.value


class ExceptionIdentifier(Identifier):

    key = 'id_Except'

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        assert node.rchild
        assert node.rchild.lchild
        assert node.rchild.lchild.value
        return node.rchild.lchild.value.value


class ReturnTypeIdentifier(Identifier):

    key = 'id_ReturnType'

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        assert node.lchild
        assert node.lchild.value
        return node.lchild.value.value


class YieldTypeIdentifier(Identifier):

    key = 'id_YieldType'

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        assert node.lchild
        assert node.lchild.value
        return node.lchild.value.value


class NoqaIdentifier(Identifier):

    key = 'id_Noqa'

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        if node.rchild and node.rchild.rchild:
            if node.rchild.rchild.value:
                return node.rchild.rchild.value.value
            elif node.rchild.rchild.lchild:
                assert node.rchild.rchild.lchild.value
                return node.rchild.rchild.lchild.value.value
            else:
                assert False
        else:
            return ''

    @staticmethod
    def extract_targets(node):
        # type: (CykNode) -> List[str]
        if (
            node.rchild
            and node.rchild.rchild
        ):
            children = list()
            for child in node.rchild.rchild.walk():
                if child.value and child.value.token_type == TokenType.WORD:
                    children.append(child.value.value)
            if len(children) > 1:
                return children[1:]
        return []



import ast
from collections import (
    deque,
)
from typing import (
    Dict,
    Optional,
    Set,
)
from ..config import get_logger


logger = get_logger()


class Context(object):
    """A context which tracks exceptions and symbols."""

    def __init__(self):
        # type: () -> None
        self.exceptions = set()  # type: Set[str]

        # A lookup from variable names to AST nodes.
        # If the variable name occurs in a raise expression,
        # then the exception will be added using this lookup.
        self.variables = dict()  # type: Dict[str, str]

        # The error(s) which the current exception block is
        # handling. (Since we only handle one handler at a time
        # in the context, and since they don't repeat the
        # exception, it's fine to overwrite this value.)
        self.handling = None  # type: Optional[str]

    def _get_name_name(self, name):
        # type: (ast.Name) -> str
        return name.id

    def _get_exception_name(self, raises):
        # type: (ast.Raise) -> str
        if isinstance(raises, str):
            return raises
        if isinstance(raises.exc, ast.Name):
            # TODO: The name could come from a higher context --
            # we should check all parent contexts.
            name = raises.exc.id
            # The name was stored in the except statement,
            # so we should restore the type.
            if name in self.variables:
                return self.variables[name]
            else:
                return name
        elif isinstance(raises.exc, ast.Call):
            if hasattr(raises.exc.func, 'id'):
                return getattr(raises.exc.func, 'id')
            elif hasattr(raises.exc.func, 'attr'):
                return getattr(raises.exc.func, 'attr')
            else:
                logger.debug(
                    'Raises function call has neither id nor attr.'
                    'has only: %s' % str(dir(raises.exc.func))
                )
        elif isinstance(raises.exc, ast.Attribute):
            return raises.exc.attr
        elif isinstance(raises.exc, ast.Subscript):
            id_repr = ''
            if hasattr(raises.exc.value, 'id'):
                id_repr = getattr(raises.exc.value, 'id')
            n_repr = ''
            if hasattr(raises.exc.slice, 'value'):
                value = getattr(raises.exc.slice, 'value')
                if hasattr(value, 'n'):
                    n_repr = getattr(value, 'n')
            return '{}[{}]'.format(
                id_repr,
                n_repr,
            )
        elif raises.exc is None:
            return self.handling or ''
        else:
            logger.debug('Unexpected type in raises expression: {}'.format(
                raises.exc
            ))
        return ''

    def add_exception(self, node):
        # type: (ast.Raise) -> None
        self.exceptions.add(self._get_exception_name(node))

    def remove_exception(self, node):
        # type: (ast.Raise) -> None
        name = self._get_exception_name(node)
        if name in self.exceptions:
            self.exceptions.remove(name)
            self.handling = name

    def remove_all_exceptions(self):
        # type: () -> None
        self.exceptions = set()

    def add_variable(self, variable, exception):
        # type: (str, ast.Name) -> None
        self.variables[variable] = self._get_name_name(exception)

    def remove_variable(self, variable):
        # type: (str) -> None
        del self.variables[variable]

    def extend(self, other):
        # type: (Context) -> None
        self.exceptions |= other.exceptions

    def finish_handling(self):
        # type: () -> None
        self.handling = None


class RaiseVisitor(ast.NodeVisitor):

    def __init__(self):
        # The context in which an exception can be raised.
        # The default context is the function body,
        # and a new context is created for each try-except
        # statement.  When the current context is finished,
        # its errors are merged upwards.
        self.contexts = deque([Context()])

    @property
    def exceptions(self):
        # type: () -> Set[str]
        return self.contexts[0].exceptions

    @property
    def context(self):
        # type: () -> Context
        return self.contexts[-1]

    def visit_Raise(self, node):
        # type: (ast.Raise) -> ast.AST
        self.context.add_exception(node)
        return self.generic_visit(node)

    def visit_Try(self, node):
        # type: (ast.Try) -> None
        self.contexts.append(Context())
        for child in node.body:
            self.visit(child)
        for handler in node.handlers:
            if handler.type:
                if handler.name and isinstance(handler.type, ast.Name):
                    self.context.add_variable(handler.name, handler.type)
                id = getattr(handler.type, 'id', None)
                if id:
                    self.context.remove_exception(id)
            else:
                self.context.remove_all_exceptions()
        for handler in node.handlers:
            self.generic_visit(handler)

            # We need to signal that we've finished handling
            # the given handler section, otherwise the caught
            # error could bleed over into a bare except clause.
            self.context.finish_handling()
        context = self.contexts.pop()
        self.context.extend(context)
